#include <apd.h>
#include <apd_internal.h>

void _apd_push_call_description(struct apd_object_struct *apd_object_internal, int importance, char *file,
      int line, const char *new_function_name, char *fmt, ...)
{
va_list argp;
char adjusted_function_name[APD_MAX_STRING_SIZE];
char description_string_temp[APD_MAX_STRING_SIZE];
char description_string_temp2[APD_MAX_STRING_SIZE];
char description_string_temp3[APD_MAX_STRING_SIZE];
struct apd_function_description_struct *function_info;

APD_INTERNAL_REPORT(5, "starting");
_apd_check_if_function_tracking_initialized(apd_object_internal);
APD_INTERNAL_ASSERT((importance >= -1*APD_IMPORTANCE_ABSOLUTE_LIMIT) && (importance <= APD_IMPORTANCE_ABSOLUTE_LIMIT),
      "(from %s:%d) importance must be with the range %d to %d.  importance = %d",
      file, line,
      -1*APD_IMPORTANCE_ABSOLUTE_LIMIT, APD_IMPORTANCE_ABSOLUTE_LIMIT, importance);
APD_INTERNAL_ASSERT(new_function_name != NULL, "");
APD_INTERNAL_ASSERT(fmt != NULL, "");
va_start(argp, fmt);
APD_INTERNAL_REPORT(5, "apd_object_internal at $%lX", (unsigned long int)apd_object_internal);
APD_INTERNAL_REPORT(5, "apd_object_internal->description_stack_size = %d", apd_object_internal->description_stack_size);

// Check to make sure main apd_object contents still match backup copy like they should
APD_CHECK_DEBUG_OBJECT(apd_object_internal);

// Check to make sure APD_FUNCTION_NAME was defined
if (strcmp(new_function_name, "!") == 0) APD_DIE("_apd_pop_call_description(): fatal error: APD_FUNCTION_NAME not defined\n");

// Allocate description structure
function_info = (struct apd_function_description_struct *)malloc( sizeof(struct apd_function_description_struct) );
APD_INTERNAL_REPORT(5, "new function_info at $%lX", (unsigned long int)function_info);
APD_INTERNAL_ASSERT(function_info != NULL, "error allocating memory for function_info");

// Create description string
snprintf(adjusted_function_name, APD_MAX_STRING_SIZE-2, "%s%s",
      apd_object_internal->level_spacers[apd_object_internal->description_stack_size], 
      new_function_name);
//  snprintf(description_string_temp, APD_MAX_STRING_SIZE/2,
//  FILE_LINE_FUNCTION_FORMAT "(start)", file, line,
//  adjusted_function_name);
snprintf(description_string_temp, APD_MAX_STRING_SIZE/3-1, APD_FUNCTION_NAME_FORMAT "(start) {{{%d ", 
         adjusted_function_name, apd_object_internal->description_stack_size+1);
vsnprintf(description_string_temp2, APD_MAX_STRING_SIZE/3-1, fmt, argp);
snprintf(description_string_temp3, APD_MAX_STRING_SIZE/3-1, APD_BEFORE_FILE_LINE_FORMAT  APD_FILE_LINE_FORMAT, file, line);

// Combine description string
strncat(description_string_temp, description_string_temp2, APD_MAX_STRING_SIZE/3-1);
strncat(description_string_temp, description_string_temp3, APD_MAX_STRING_SIZE/3-1);

// Allocate memory for the description string
function_info->description = (char *)malloc(strlen(description_string_temp)+1);
APD_INTERNAL_REPORT(5, "new function_info->description at $%lX", (unsigned long int)function_info->description);
APD_INTERNAL_REPORT(5, "function_info->description allocated %d bytes at $%lX",
      strlen(description_string_temp)+1, (unsigned long int)function_info->description);
APD_INTERNAL_ASSERT(function_info->description != NULL, "error allocating memory for function_info->description");
APD_INTERNAL_REPORT(5, "copying description string into function_info->description");
strcpy(function_info->description, description_string_temp);

// Decrease report_threshold if this option is set
if (apd_object_internal->report_threshold_decreases_with_stack == 1) apd_object_internal->report_threshold--;

// Set importance for this function
function_info->importance = importance;

// Allocate memory for the name string
function_info->name = (char *)malloc(strlen(adjusted_function_name)+1);
APD_INTERNAL_REPORT(5, "new function_info->name at $%lX", (unsigned long int)function_info->name);
APD_INTERNAL_REPORT(5, "function_info->name allocated %d bytes at $%lX",
      strlen(new_function_name)+1, (unsigned long int)function_info->name);
APD_INTERNAL_ASSERT(function_info->name != NULL, "error allocating memory for function_info->name");
strcpy(function_info->name, adjusted_function_name);

// point to current top function name
apd_object_internal->current_function_name = function_info->name;

// If report threshold high enough, print this entry to the error stream
if (apd_object_internal->report_threshold >= function_info->importance)
  {
    fprintf(apd_object_internal->error_stream, "%s\n", function_info->description);
    fflush(apd_object_internal->error_stream);  // Make sure it is printed in case of a crash
  }

// Insert description_string into description call stack
APD_INTERNAL_REPORT(3, "inserting function_info into description stack");
apd_object_internal->description_stack = g_slist_insert(apd_object_internal->description_stack, (gpointer)function_info, 0);
apd_object_internal->description_stack_size++;

// Check for overflowing description stack caused by a function having
//   a PUSH not matched with a POP
if (apd_object_internal->description_stack_size > apd_object_internal->max_function_depth)
  {
    function_info = (struct apd_function_description_struct *)
            g_slist_nth_data(apd_object_internal->description_stack, apd_object_internal->max_function_depth-5);
    fprintf(apd_object_internal->error_stream, "_apd_push_call_description(): Description stack seems to be overflowing.\n");
    fprintf(apd_object_internal->error_stream, "  A good hint at which function is causing this is to look at the description\n");
    fprintf(apd_object_internal->error_stream, "  a few items back on the stack: \n%s\n", function_info->description);
    exit(EXIT_FAILURE);
  }
va_end(argp);

// Refresh apd_object's backup copy since we just changed it
APD_BACKUP_DEBUG_OBJECT(apd_object_internal);

APD_INTERNAL_REPORT(5, "ending");
} // end _apd_push_call_description
